# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

# * The software is provided "as is" without any warranty, neither express
#   nor implied.
# * In no event will the author be held liable for any damage(s) arising
#   from the use of the software.
# * Redistribution of the software is allowed only in unmodified form.
# * Permission is granted to use the software solely for personal private
#   and non-commercial purposes.
# * An individuals use of the software in his or her capacity or function
#   as an agent, (independent) contractor, employee, member or officer of
#   a business, corporation or organization (commercial or non-commercial)
#   does not qualify as personal private and non-commercial purpose.
# * Without written approval from the author the software must not be used
#   for a business, for commercial, corporate, governmental, military or
#   organizational purposes of any kind, or in a commercial, corporate,
#   governmental, military or organizational environment of any kind.

if [ $# -eq 1 -a "$1" = 'sha-ni' ]
then
as --32 --defsym ALIGNED=yes --no-pad-sections --noexecstack -o sha2-256.o <<+\E\O\F
# Copyright  2017-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.ident	"Copyright (C) 2017-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>"
.file	"sha2-256-ni.s"
.arch	znver1
.code32
.att_syntax

.struct	0				# SHA256_CTX structure
state:					# 8 double words: {h, g, d, c, f, e, b, a}
.space	32
count:					# 2 double words
.space	8
block:					# 16 double words
.space	64

.section .rodata, "ar", @progbits
.align	64
reverse:				# constants for endian conversion
.octa	0x000102030405060708090A0B0C0D0E0F
endian:
.long	0x00010203, 0x04050607, 0x08090A0B, 0x0C0D0E0F
hgdc:					# SHA-256 start values: H7, H6, H3, H2
.long	0x5BE0CD19, 0x1F83D9AB, 0xA54FF53A, 0x3C6EF372
feba:					# SHA-256 start values: H5, H4, H1, H0
.long	0x9B05688C, 0x510E527F, 0xBB67AE85, 0x6A09E667
k256:					# SHA-256 round constants
.long	0x428A2F98, 0x71374491, 0xB5C0FBCF, 0xE9B5DBA5
.long	0x3956C25B, 0x59F111F1, 0x923F82A4, 0xAB1C5ED5
.long	0xD807AA98, 0x12835B01, 0x243185BE, 0x550C7DC3
.long	0x72BE5D74, 0x80DEB1FE, 0x9BDC06A7, 0xC19BF174
.long	0xE49B69C1, 0xEFBE4786, 0x0FC19DC6, 0x240CA1CC
.long	0x2DE92C6F, 0x4A7484AA, 0x5CB0A9DC, 0x76F988DA
.long	0x983E5152, 0xA831C66D, 0xB00327C8, 0xBF597FC7
.long	0xC6E00BF3, 0xD5A79147, 0x06CA6351, 0x14292967
.long	0x27B70A85, 0x2E1B2138, 0x4D2C6DFC, 0x53380D13
.long	0x650A7354, 0x766A0ABB, 0x81C2C92E, 0x92722C85
.long	0xA2BFE8A1, 0xA81A664B, 0xC24B8B70, 0xC76C51A3
.long	0xD192E819, 0xD6990624, 0xF40E3585, 0x106AA070
.long	0x19A4C116, 0x1E376C08, 0x2748774C, 0x34B0BCB5
.long	0x391C0CB3, 0x4ED8AA4A, 0x5B9CCA4F, 0x682E6FF3
.long	0x748F82EE, 0x78A5636F, 0x84C87814, 0x8CC70208
.long	0x90BEFFFA, 0xA4506CEB, 0xBEF9A3F7, 0xC67178F2
.size	k256, .-k256
.type	k256, @object

.ifnotdef ORIGINAL
.macro	sha256	m0 :req, m1 :req, m2 :req, m3 :req
.if \@ < 4
	movdqu	block+16*\@(%edx), \m0
	pshufb	%xmm3, \m0		# \m0 = {ntohl(block[\@*16]), ntohl(block[\@*16+1]),
					#        ntohl(block[\@*16+2]), ntohl(block[\@*16+3])}
					#     = {w[\@*16], w[\@*16+1], w[\@*16+2], w[\@*16+3]}
.else
					# \m0 = {w[\@*16-16], w[\@*16-15], w[\@*16-14], w[\@*16-13]},
					# \m1 = {w[\@*16-12], w[\@*16-11], w[\@*16-10], w[\@*16-9]},
					# \m2 = {w[\@*16-8],  w[\@*16-7],  w[\@*16-6],  w[\@*16-5]},
					# \m3 = {w[\@*16-4],  w[\@*16-3],  w[\@*16-2],  w[\@*16-1]}
	sha256msg1 \m1, \m0
	movdqa	\m3, %xmm0
	palignr	$4, \m2, %xmm0		# xmm0 = {w[\@*16-7], w[\@*16-6], w[\@*16-5], w[\@*16-4]}
	paddd	%xmm0, \m0
	sha256msg2 \m3, \m0		# \m0 = {w[\@*16], w[\@*16+1], w[\@*16+2], w[\@*16+3]}
.endif
	movdqa	\@*16-128(%ecx), %xmm0	# xmm0 = {k[\@*16], k[\@*16+1], k[\@*16+2], k[\@*16+3]}
	paddd	\m0, %xmm0		# xmm0 = {(w+k)[\@*16], (w+k)[\@*16+1], (w+k)[\@*16+2], (w+k)[\@*16+3]}
	sha256rnds2 %xmm1, %xmm2	# xmm2 = {f', e', b', a'},
					# xmm1 = {h', g', d', c'}
	punpckhqdq %xmm0, %xmm0		# xmm0 = {(w+k)[\@*16+2], (w+k)[\@*16+3], (w+k)[\@*16+2], (w+k)[\@*16+3]}
	sha256rnds2 %xmm2, %xmm1	# xmm1 = {h", g", d", c"},
					# xmm2 = {f", e", b", a"}
.endm
.else # ORIGINAL
.macro	sha256	m0 :req, m1 :req, m2 :req, m3 :req
.if \@ < 4
	movdqu	block+16*\@(%edx), \m0
	pshufb	%xmm3, \m0
.endif
	movdqa	\@*16-128(%ecx), %xmm0	# xmm0 = {k[\@*16], k[\@*16+1], k[\@*16+2], k[\@*16+3]}
	paddd	\m0, %xmm0		# xmm0 = {(w+k)[\@*16], (w+k)[\@*16+1], (w+k)[\@*16+2], (w+k)[\@*16+3]}
	sha256rnds2 %xmm2, %xmm1	# xmm1 = {h', g', d', c'},
					# xmm2 = {f', e', b', a'}
.if \@ > 2 && \@ < 15
	movdqa	\m0, %xmm3
	palignr	$4, \m3, %xmm3		# xmm3 = {(w+k)[\@*16-3], (w+k)[\@*16-2], (w+k)[\@*16-1], (w+k)[\@*16]}
	paddd	%xmm3, \m1
	sha256msg2 \m0, \m1
.endif
	punpckhqdq %xmm0, %xmm0
	sha256rnds2 %xmm1, %xmm2	# xmm2 = {f", e", b", a"},
					# xmm1 = {h", g", d", c"}
.if \@ > 0 && \@ < 13
	sha256msg1 \m0, \m3
.endif
.endm
.endif # ORIGINAL
.text

sha256_core:				# void SHA256_Core(SHA256_CTX *context)

	leal	k256+128, %ecx		# ecx = address of round constants
	movl	4(%esp), %edx		# edx = address of context structure
	movdqu	state(%edx), %xmm1	# xmm1 = {h, g, d, c}
	movdqu	state+16(%edx), %xmm2	# xmm2 = {f, e, b, a}
	movdqa	endian, %xmm3		# xmm3 = constant for endian conversion
.rept 4					# 4*4*4 rounds
	sha256	%xmm4, %xmm5, %xmm6, %xmm7
	sha256	%xmm5, %xmm6, %xmm7, %xmm4
	sha256	%xmm6, %xmm7, %xmm4, %xmm5
	sha256	%xmm7, %xmm4, %xmm5, %xmm6
.endr
	movdqu	state(%edx), %xmm6	# xmm6 = {h, g, d, c}
	movdqu	state+16(%edx), %xmm7	# xmm7 = {f, e, b, a}
	paddd	%xmm6, %xmm1		# xmm1 = {h', g', d', c'} + {h, g, d, c}
					#      = {h", g", d", c"}
	paddd	%xmm7, %xmm2		# xmm2 = {f', e', b', a'} + {f, e, b, a}
					#      = {f", e", b", a"}
	movdqu	%xmm1, state(%edx)
	movdqu	%xmm2, state+16(%edx)	# state = {h", g", d", c", f", e", b", a"}
	retl

.global	sha256_core
.size	sha256_core, .-sha256_core
.type	sha256_core, @function

sha256_final:				# void SHA256_Final(SHA256_CTX    *context,
					#                   unsigned char digest[32])

	movl	4(%esp), %edx		# edx = address of context structure
	movl	count(%edx), %ecx	# ecx = low double word of count
	andl	$63, %ecx		# ecx = number of bytes in block
					#     = index of first free byte in block
	pushl	%edi
	leal	block(%edx, %ecx), %edi	# edi = address of first free byte in block
	pushl	%edi
.Lpad_1:
	movb	$0b10000000, %al
	stosb				# block[index] = 0b10000000
.Lpad_0:
	xorl	%eax, %eax		# eax = 0
	xorl	$63, %ecx		# ecx = number of free bytes in block - 1
					#     = 63 - index
	rep	stosb			# block[index + 1, 63] = 0
	subl	$8, %edi		# edi = address of last quad word in block
	popl	%eax			# eax = address of first free byte in block
	cmpl	%eax, %edi
	ja	.Lpad_count		# space for count available in block?
					# index < 56?
	pushl	%edx
	calll	sha256_core
	popl	%edx
.Lpad_block:
	movl	%edi, %ecx		# ecx = address of last quad word in block
	leal	block(%edx), %edi	# edi = address of block
	xorl	%eax, %eax		# eax = 0
	subl	%edi, %ecx		# ecx = number of bytes before last quad word
					#     = 56
	rep	stosb			# block[0, 55] = 0,
					# edi = address of last quad word in block
.Lpad_count:
	movl	count(%edx), %ecx
	movl	count+4(%edx), %eax	# eax:ecx = count
	shldl	$3, %ecx, %eax
	shll	$3, %ecx		# eax:ecx = count * 8
					#         = number of message bits
	bswapl	%eax
	bswapl	%ecx			# eax:ecx = htonll(number of message bits)
	stosl
	movl	%ecx, %eax
	stosl				# block[56, 63] = htonll(number of message bits)
	pushl	%edx
	calll	sha256_core
	popl	%edx
.Ldigest:
	movdqu	state(%edx), %xmm1	# xmm1 = {h, g, d, c}
	movdqu	state+16(%edx), %xmm2	# xmm2 = {f, e, b, a}
	movdqa	%xmm1, %xmm0		# xmm0 = {h, g, d, c}
	punpcklqdq %xmm2, %xmm1		# xmm1 = {h, g, f, e}
	punpckhqdq %xmm2, %xmm0		# xmm0 = {d, c, b, a}
	movdqa	reverse, %xmm2		# xmm2 = shuffle indices
	pshufb	%xmm2, %xmm1		# xmm1 = {htonl(e), htonl(f), htonl(g), htonl(h)}
	pshufb	%xmm2, %xmm0		# xmm0 = {htonl(a), htonl(b), htonl(c), htonl(d)}
	movl	12(%esp), %edx		# edx = address of digest
	movdqu	%xmm0, (%edx)
	movdqu	%xmm1, 16(%edx)
	popl	%edi
	retl

.global	sha256_final
.size	sha256_final, .-sha256_final
.type	sha256_final, @function

sha256_init:				# void SHA256_Init(SHA256_CTX *context)

	movl	4(%esp), %edx		# edx = address of context structure
	xorl	%eax, %eax		# eax = 0
	movdqa	hgdc, %xmm1		# xmm1 = {H7, H6, H3, H2}
	movdqa	feba, %xmm2		# xmm2 = {H5, H4, H1, H0}
	movl	%eax, count(%edx)
	movl	%eax, count+4(%edx)	# count = 0
	movdqu	%xmm1, state(%edx)
	movdqu	%xmm2, state+16(%edx)	# state = {H7, H6, H3, H2, H5, H4, H1, H0}
					#       = {h, g, d, c, f, e, b, a}
	retl

.global	sha256_init
.size	sha256_init, .-sha256_init
.type	sha256_init, @function

sha256_update:				# void SHA256_Update(SHA256_CTX   *context,
					#                    void const   *data,
					#                    unsigned int size)

	movl	12(%esp), %eax		# eax = number of bytes in data
	testl	%eax, %eax
	jz	.Lnone			# no data?

	movl	4(%esp), %edx		# edx = address of context structure
	movl	count(%edx), %ecx	# ecx = low double word of count
	andl	$63, %ecx		# ecx = number of bytes in block
					#     = index of first free byte in block
	addl	%eax, count(%edx)
	adcl	$0, count+4(%edx)	# count += number of bytes in data
	pushl	%esi
	movl	12(%esp), %esi		# esi = address of data
	pushl	%edi
.Ldata:
	leal	block(%edx, %ecx), %edi	# edi = address of first free byte in block
	xorl	$63, %ecx
	incl	%ecx			# ecx = number of free bytes in block
	subl	%ecx, %eax		# eax = number of bytes in data
					#     - number of free bytes in block
	jb	.Llast			# number of bytes in data < number of free bytes in block?
.Lmore:
	rep	movsb			# esi = address of remaining data,
					# ecx = 0 = index of first free byte in block
	pushl	%edx
	calll	sha256_core
	popl	%edx
	testl	%eax, %eax
	jnz	.Ldata			# more data?

	popl	%edi
	popl	%esi
.Lnone:
	retl
.Llast:
	addl	%eax, %ecx		# ecx = number of bytes in data
	rep	movsb
	popl	%edi
	popl	%esi
	retl

.global	sha256_update
.size	sha256_update, .-sha256_update
.type	sha256_update, @function
.end
+EOF
else
as --32 --defsym ALIGNED=yes --no-pad-sections --noexecstack -o sha2-256.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

# Common "cdecl" calling and naming convention for i386 platform:
# - arguments are pushed on stack in reverse order (from right to left),
#   4-byte aligned;
# - 64-bit integer arguments are passed as pair of 32-bit integer arguments,
#   low part below high part;
# - 64-bit integer result is returned in registers EAX (low part) and
#   EDX (high part);
# - 32-bit integer or pointer result is returned in register EAX;
# - registers EAX, ECX and EDX are volatile and can be clobbered;
# - registers EBX, ESP, EBP, ESI and EDI must be preserved;
# - function names are prefixed with an underscore.

.ident	"Copyright (C) 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>"
.file	"sha2-256.s"
.arch	i486				# support BSWAP instruction
.code32
.att_syntax

.struct	0				# SHA256_CTX structure
state:					# 8 double words
.space	8*4
count:					# 1 quad word
.space	2*4
block:					# 16 double words
.space	16*4

.macro	sha256	k :req, t :req
.if \t < 16
	movl	block+4*\t(%ebp), %eax
	bswapl	%eax			# eax = ntohl(block[t])
.else
	movl	block+4*((\t-15)&15)(%ebp), %eax
	movl	block+4*((\t-2)&15)(%ebp), %ebx
	movl	%eax, %ecx		# eax = m = block[t-15&15]
	movl	%ebx, %edx		# ebx = n = block[t-2&15]
	shrl	$3, %eax		# eax = m >> 3
	shrl	$10, %ebx		# ebx = n >> 10
	rorl	$7, %ecx		# ecx = ROTR(m, 7)
	rorl	$17, %edx		# edx = ROTR(n, 17)
	xorl	%ecx, %eax		# eax = (m >> 3) ^ ROTR(m, 7)
	xorl	%edx, %ebx		# ebx = (n >> 10) ^ ROTR(n, 17)
	rorl	$(18-7), %ecx		# ecx = ROTR(m, 18)
	rorl	$(19-17), %edx		# edx = ROTR(n, 19)
	xorl	%ecx, %eax		# eax = (m >> 3) ^ ROTR(m, 7) ^ ROTR(m, 18)
					#     = SMALL_0(m)
	xorl	%edx, %ebx		# ebx = (n >> 10) ^ ROTR(n, 17) ^ ROTR(n, 19)
					#     = SMALL_1(n)
	addl	block+4*((\t-16)&15)(%ebp), %eax
	addl	block+4*((\t-7)&15)(%ebp), %ebx
	addl	%ebx, %eax		# eax = SMALL_0(block[t-15&15]) + block[t-16&15]
					#     + SMALL_1(block[t-2&15]) + block[t-7&15]
.endif
	movl	%eax, block+4*((\t)&15)(%ebp)
					# block[t&15] = (t < 16)
					#             ? ntohl(block[t])
					#             : SMALL_0(block[t-15&15]) + block[t-16&15]
					#             + SMALL_1(block[t-2&15]) + block[t-7&15]
	addl	state+4*((71-\t)&7)(%ebp), %eax
					# eax = block[t&15] + state[71-t&7]
					#     = block[t&15] + h
	movl	state+4*((70-\t)&7)(%ebp), %ebx
					# ebx = g = state[70-t&7]
	movl	state+4*((69-\t)&7)(%ebp), %ecx
					# ecx = f = state[69-t&7]
	movl	state+4*((68-\t)&7)(%ebp), %edx
					# edx = e = state[68-t&7]
	xorl	%ebx, %ecx		# ecx = f ^ g
	andl	%edx, %ecx		# ecx = e & (f ^ g)
	xorl	%ebx, %ecx		# ecx = e & (f ^ g) ^ g
					#     = CH(e, f, g)
	addl	%ecx, %eax		# eax = block[t&15] + CH(e, f, g) + h
	movl	%edx, %ebx		# ebx = e
	rorl	$6, %edx		# edx = ROTR(e, 6)
	rorl	$11, %ebx		# ebx = ROTR(e, 11)
	xorl	%ebx, %edx		# edx = ROTR(e, 6) ^ ROTR(e, 11)
	rorl	$(25-11), %ebx		# ebx = ROTR(e, 25)
	xorl	%edx, %ebx		# ebx = ROTR(e, 6) ^ ROTR(e, 11) ^ ROTR(e, 25)
					#     = SIGMA_1(e)
	leal	\k(%eax, %ebx), %eax	# eax = SIGMA_1(e) + CH(e, f, g) + h + k[t] + block[t&15]
					#     = T1
	movl	state+4*((64-\t)&7)(%ebp), %ebx
	movl	state+4*((65-\t)&7)(%ebp), %ecx
					# ecx = b
	movl	%ebx, %edx		# edx = a
	xorl	%ecx, %ebx		# ebx = a ^ b
	andl	%edx, %ecx		# ecx = a & b
	andl	state+4*((66-\t)&7)(%ebp), %ebx
					# ebx = (a ^ b) & c
	addl	%eax, state+4*((67-\t)&7)(%ebp)
					# d' = d + T1
	orl	%ecx, %ebx		# ebx = (a & b) | ((a ^ b) & c)
					#     = MAJ(a, b, c)
	addl	%ebx, %eax		# eax = T1 + MAJ(a, b, c)
	movl	%edx, %ecx		# ecx = a
	rorl	$2, %edx		# edx = ROTR(a, 2)
	rorl	$13, %ecx		# ecx = ROTR(a, 13)
	xorl	%ecx, %edx		# edx = ROTR(a, 2) ^ ROTR(a, 13)
	rorl	$(22-13), %ecx		# ecx = ROTR(a, 22)
	xorl	%ecx, %edx		# edx = ROTR(a, 2) ^ ROTR(a, 13) ^ ROTR(a, 22)
					#     = SIGMA_0(a)
	addl	%edx, %eax		# eax = T1 + T2
	movl	%eax, state+4*((63-\t)&7)(%ebp)
					# h' = T1 + T2
.endm
.text

sha256_core:				# void SHA256_Core(SHA256_CTX *context)

	pushl	%ebp
	movl	8(%esp), %ebp		# ebp = address of context structure
	pushl	%ebx
					# save state
	pushl	state(%ebp)
	pushl	state+4(%ebp)
	pushl	state+8(%ebp)
	pushl	state+12(%ebp)
	pushl	state+16(%ebp)
	pushl	state+20(%ebp)
	pushl	state+24(%ebp)
	pushl	state+28(%ebp)
					# calculate 64 rounds
	sha256	0x428A2F98, 0
	sha256	0x71374491, 1
	sha256	0xB5C0FBCF, 2
	sha256	0xE9B5DBA5, 3
	sha256	0x3956C25B, 4
	sha256	0x59F111F1, 5
	sha256	0x923F82A4, 6
	sha256	0xAB1C5ED5, 7
	sha256	0xD807AA98, 8
	sha256	0x12835B01, 9
	sha256	0x243185BE, 10
	sha256	0x550C7DC3, 11
	sha256	0x72BE5D74, 12
	sha256	0x80DEB1FE, 13
	sha256	0x9BDC06A7, 14
	sha256	0xC19BF174, 15
	sha256	0xE49B69C1, 16
	sha256	0xEFBE4786, 17
	sha256	0x0FC19DC6, 18
	sha256	0x240CA1CC, 19
	sha256	0x2DE92C6F, 20
	sha256	0x4A7484AA, 21
	sha256	0x5CB0A9DC, 22
	sha256	0x76F988DA, 23
	sha256	0x983E5152, 24
	sha256	0xA831C66D, 25
	sha256	0xB00327C8, 26
	sha256	0xBF597FC7, 27
	sha256	0xC6E00BF3, 28
	sha256	0xD5A79147, 29
	sha256	0x06CA6351, 30
	sha256	0x14292967, 31
	sha256	0x27B70A85, 32
	sha256	0x2E1B2138, 33
	sha256	0x4D2C6DFC, 34
	sha256	0x53380D13, 35
	sha256	0x650A7354, 36
	sha256	0x766A0ABB, 37
	sha256	0x81C2C92E, 38
	sha256	0x92722C85, 39
	sha256	0xA2BFE8A1, 40
	sha256	0xA81A664B, 41
	sha256	0xC24B8B70, 42
	sha256	0xC76C51A3, 43
	sha256	0xD192E819, 44
	sha256	0xD6990624, 45
	sha256	0xF40E3585, 46
	sha256	0x106AA070, 47
	sha256	0x19A4C116, 48
	sha256	0x1E376C08, 49
	sha256	0x2748774C, 50
	sha256	0x34B0BCB5, 51
	sha256	0x391C0CB3, 52
	sha256	0x4ED8AA4A, 53
	sha256	0x5B9CCA4F, 54
	sha256	0x682E6FF3, 55
	sha256	0x748F82EE, 56
	sha256	0x78A5636F, 57
	sha256	0x84C87814, 58
	sha256	0x8CC70208, 59
	sha256	0x90BEFFFA, 60
	sha256	0xA4506CEB, 61
	sha256	0xBEF9A3F7, 62
	sha256	0xC67178F2, 63
					# add saved state
	popl	%eax
	addl	%eax, state+28(%ebp)
	popl	%ebx
	addl	%ebx, state+24(%ebp)
	popl	%ecx
	addl	%ecx, state+20(%ebp)
	popl	%edx
	addl	%edx, state+16(%ebp)
	popl	%eax
	addl	%eax, state+12(%ebp)
	popl	%ebx
	addl	%ebx, state+8(%ebp)
	popl	%ecx
	addl	%ecx, state+4(%ebp)
	popl	%edx
	addl	%edx, state(%ebp)
	popl	%ebx
	popl	%ebp
	retl

.global	sha256_core
.size	sha256_core, .-sha256_core
.type	sha256_core, @function

sha256_final:				# void SHA256_Final(SHA256_CTX    *context,
					#                   unsigned char digest[32])

	movl	4(%esp), %edx		# edx = address of context structure
	movl	count(%edx), %ecx	# ecx = low double word of count
	andl	$63, %ecx		# ecx = number of bytes in block
					#     = index of first free byte in block
	pushl	%edi
	leal	block(%edx, %ecx), %edi	# edi = address of first free byte in block
	pushl	%edi
.Lpad_1:
	movb	$0b10000000, %al
	stosb				# block[index] = 0b10000000
.Lpad_0:
	xorl	%eax, %eax		# eax = 0
	xorl	$63, %ecx		# ecx = number of free bytes in block - 1
					#     = 63 - index
	rep	stosb			# block[index + 1, 63] = 0
	subl	$8, %edi		# edi = address of last quad word in block
	popl	%eax			# eax = address of first free byte in block
	cmpl	%eax, %edi
	ja	.Lpad_count		# space for count available in block?
					# index < 56?
	pushl	%edx
	calll	sha256_core
	popl	%edx
.Lpad_block:
	movl	%edi, %ecx		# ecx = address of last quad word in block
	leal	block(%edx), %edi	# edi = address of block
	xorl	%eax, %eax		# eax = 0
	subl	%edi, %ecx		# ecx = number of bytes before last quad word
					#     = 56
	rep	stosb			# block[0, 55] = 0,
					# edi = address of last quad word in block
.Lpad_count:
	movl	count(%edx), %ecx
	movl	count+4(%edx), %eax	# eax:ecx = count
	shldl	$3, %ecx, %eax
	shll	$3, %ecx		# eax:ecx = count * 8
					#         = number of message bits
	bswapl	%eax
	bswapl	%ecx			# eax:ecx = htonll(number of message bits)
	stosl
	movl	%ecx, %eax
	stosl				# block[56, 63] = htonll(number of message bits)
	pushl	%edx
	calll	sha256_core
	popl	%edx
	xchgl	%edx, %esi		# esi = address of state
	movl	12(%esp), %edi		# edi = address of digest
	movl	$8, %ecx		# ecx = number of double words
.Ldigest:
	lodsl
.ifdef ALIGNED
	bswapl	%eax
	stosl
.else
	roll	$8, %eax
	stosb
	roll	$8, %eax
	stosb
	roll	$8, %eax
	stosb
	roll	$8, %eax
	stosb
.endif
	decl	%ecx
	jnz	.Ldigest

	movl	%edx, %esi
	popl	%edi
	retl

.global	sha256_final
.size	sha256_final, .-sha256_final
.type	sha256_final, @function

sha256_init:				# void SHA256_Init(SHA256_CTX *context)

	movl	%edi, %edx
	movl	4(%esp), %edi		# edi = address of context structure
	movl	0x6A09E667, %eax	# eax = H0
	stosl				# state[0] = H0
	movl	0xBB67AE85, %eax	# eax = H1
	stosl				# state[1] = H1
	movl	0x3C6EF372, %eax	# eax = H2
	stosl				# state[2] = H2
	movl	0xA54FF53A, %eax	# eax = H3
	stosl				# state[3] = H3
	movl	0x510E527F, %eax	# eax = H4
	stosl				# state[4] = H4
	movl	0x9B05688C, %eax	# eax = H5
	stosl				# state[5] = H5
	movl	0x1F83D9AB, %eax	# eax = H6
	stosl				# state[6] = H6
	movl	0x5BE0CD19, %eax	# eax = H7
	stosl				# state[7] = H7
	xorl	%eax, %eax		# eax = 0
	stosl
	stosl				# count = 0
	movl	%edx, %edi
	retl

.global	sha256_init
.size	sha256_init, .-sha256_init
.type	sha256_init, @function

sha256_update:				# void SHA256_Update(SHA256_CTX   *context,
					#                    void const   *data,
					#                    unsigned int size)

	movl	12(%esp), %eax		# eax = number of bytes in data
	testl	%eax, %eax
	jz	.Lnone			# no data?

	movl	4(%esp), %edx		# edx = address of context structure
	movl	count(%edx), %ecx	# ecx = low double word of count
	andl	$63, %ecx		# ecx = number of bytes in block
					#     = index of first free byte in block
	addl	%eax, count(%edx)
	adcl	$0, count+4(%edx)	# count += number of bytes in data
	pushl	%esi
	movl	12(%esp), %esi		# esi = address of data
	pushl	%edi
.Ldata:
	leal	block(%edx, %ecx), %edi	# edi = address of first free byte in block
	xorl	$63, %ecx
	incl	%ecx			# ecx = number of free bytes in block
	subl	%ecx, %eax		# eax = number of bytes in data
					#     - number of free bytes in block
	jb	.Llast			# number of bytes in data < number of free bytes in block?
.Lmore:
	rep	movsb			# esi = address of remaining data
	movl	%eax, %edi
	pushl	%edx
	calll	sha256_core
	popl	%edx
	movl	%edi, %eax
	xorl	%ecx, %ecx		# ecx = 0 = index of first free byte in block
	testl	%eax, %eax
	jnz	.Ldata			# more data?

	popl	%edi
	popl	%esi
.Lnone:
	retl
.Llast:
	addl	%eax, %ecx		# ecx = number of bytes in data
	rep	movsb
	popl	%edi
	popl	%esi
	retl

.global	sha256_update
.size	sha256_update, .-sha256_update
.type	sha256_update, @function
.end
+EOF
fi

gcc -m32 -O2 -x c - -x none sha2-256.o <<+\E\O\F && ./a.out
// Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

typedef struct _sha256_ctx {
    unsigned int state[8], count[2], block[16];
} sha256_ctx;

void sha256_core(sha256_ctx *context);
void sha256_final(sha256_ctx *context, unsigned char digest[32]);
void sha256_init(sha256_ctx *context);
void sha256_update(sha256_ctx *context, unsigned char const data[], unsigned int size);

#include <arpa/inet.h>          // for ntohl() function
#include <stdio.h>
#include <string.h>             // for memcpy() and memset() functions
#include <time.h>

int main()
{
    unsigned char digest[32], million[1000000];
    unsigned int  n = 1024 * 1024 * 1024 / 64;

    clock_t t;

    sha256_ctx context;

    sha256_init(&context);
    sha256_final(&context, digest);

    printf("\"\"\n"
           "\te3b0c442 98fc1c14 9afbf4c8 996fb924 27ae41e4 649b934c a495991b 7852b855\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    sha256_init(&context);
    sha256_update(&context, "abc", 3);
    sha256_final(&context, digest);

    printf("\"abc\"\n"
           "\tba7816bf 8f01cfea 414140de 5dae2223 b00361a3 96177a9c b410ff61 f20015ad\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    sha256_init(&context);
    sha256_update(&context, "abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
                     sizeof("abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq") - 1);
    sha256_final(&context, digest);

    printf("\"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq\"\n"
           "\t248d6a61 d20638b8 e5c02693 0c3e6039 a33ce459 64ff2167 f6ecedd4 19db06c1\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    sha256_init(&context);
    sha256_update(&context, "abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu",
                     sizeof("abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu") - 1);
    sha256_final(&context, digest);

    printf("\"abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu\"\n"
           "\tcf5b16a7 78af8380 036ce59e 7b049237 0b249b11 e8f07a51 afac4503 7afee9d1\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    sha256_init(&context);
    sha256_update(&context, "\xBD", 1);
    sha256_final(&context, digest);

    printf("\"\\xBD\"\n"
           "\t68325720 aabd7c82 f30f554b 313d0570 c95accbb 7dc4b5aa e11204c0 8ffe732b\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    sha256_init(&context);
    sha256_update(&context, "\xC9\x8C\x8E\x55", 4);
    sha256_final(&context, digest);

    printf("\"\\xC9\\x8C\\x8E\\x55\"\n"
           "\t7abc22c0 ae5af26c e93dbb94 433a0e0b 2e119d01 4f8e7f65 bd56c61c cccd9504\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    memset(million, 0, 1000);

    sha256_init(&context);
    sha256_update(&context, million, 55);
    sha256_final(&context, digest);

    printf("'\\0'*55\n"
           "\t02779466 cdec1638 11d07881 5c633f21 90141308 1449002f 24aa3e80 f0b88ef7\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    sha256_init(&context);
    sha256_update(&context, million, 56);
    sha256_final(&context, digest);

    printf("'\\0'*56\n"
           "\td4817aa5 497628e7 c77e6b60 6107042b bba31308 88c5f47a 375e6179 be789fbb\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    sha256_init(&context);
    sha256_update(&context, million, 57);
    sha256_final(&context, digest);

    printf("'\\0'*57\n"
           "\t65a16cb7 861335d5 ace3c607 18b5052e 44660726 da4cd13b b745381b 235a1785\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    sha256_init(&context);
    sha256_update(&context, million, 64);
    sha256_final(&context, digest);

    printf("'\\0'*64\n"
           "\tf5a5fd42 d16a2030 2798ef6e d309979b 43003d23 20d9f0e8 ea9831a9 2759fb4b\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    sha256_init(&context);
    sha256_update(&context, million, 1000);
    sha256_final(&context, digest);

    printf("'\\0'*1000\n"
           "\t541b3e9d aa09b20b f85fa273 e5cbd3e8 0185aa4e c298e765 db87742b 70138a53\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    memset(million, 'A', 1000);

    sha256_init(&context);
    sha256_update(&context, million, 1000);
    sha256_final(&context, digest);

    printf("'A'*1000\n"
           "\tc2e68682 3489ced2 017f6059 b8b23931 8b6364f6 dcd835d0 a519105a 1eadd6e4\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    memset(million, 'U', 1005);

    sha256_init(&context);
    sha256_update(&context, million, 1005);
    sha256_final(&context, digest);

    printf("'U'*1005\n"
           "\tf4d62dde c0f3dd90 ea1380fa 16a5ff8d c4c54b21 740650f2 4afc4120 903552b0\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    memset(million, 'a', 1000000);

    sha256_init(&context);
    sha256_update(&context, million, 1000000);
    sha256_final(&context, digest);

    printf("'a'*1000000\n"
           "\tcdc76e5c 9914fb92 81a1c7e2 84d73e67 f1809a48 a497200e 046d39cc c7112cd0\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    memset(million, 0, 1000000);

    sha256_init(&context);
    sha256_update(&context, million, 1000000);
    sha256_final(&context, digest);

    printf("'\\0'*1000000\n"
           "\td29751f2 649b32ff 572b5e0a 9f541ea6 60a50f94 ff0beedf b0b692b9 24cc8025\n"
           "\t%08X %08X %08X %08X %08X %08X %08X %08X\n",
           ntohl(context.state[0]), ntohl(context.state[1]),
           ntohl(context.state[2]), ntohl(context.state[3]),
           ntohl(context.state[4]), ntohl(context.state[5]),
           ntohl(context.state[6]), ntohl(context.state[7]));

    t = clock();

    do sha256_core(&context); while (--n);

    t = clock() - t;

    printf("%lu.%06lu seconds per GiB\n",
           t / CLOCKS_PER_SEC, (t % CLOCKS_PER_SEC) * 1000000u / CLOCKS_PER_SEC);
}
+EOF

if [ $# -eq 1 -a "$1" = 'sha-ni' ]
then
as --64 --defsym ALIGNED=yes --no-pad-sections --noexecstack -o sha2-512.o <<+\E\O\F
# Copyright  2017-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.ident	"Copyright (C) 2017-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>"
.file	"sha-ni-512.s"
.arch	znver1
.code64
.att_syntax

.struct	0				# SHA512_CTX structure
state:					# 8 quad words: {h, g, d, c, f, e, b, a}
.space	64
block:					# 16 quad words
.space	128
count:					# 2 quad words
.space	16

.section .const, "ar"
.align	64
reverse:				# constants for endian conversion
.octa	0x00102030405060708090A0B0C0D0E0F
.octa	0x00102030405060708090A0B0C0D0E0F
endian:
.quad	0x0001020304050607, 0x08090A0B0C0D0E0F
.quad	0x0001020304050607, 0x08090A0B0C0D0E0F
hgdc:					# SHA-512 start values: H7, H6, H3, H2
.quad	0x5BE0CD19137E2179, 0x1F83D9ABFB41BD6B, 0xA54FF53A5F1D36F1, 0x3C6EF372FE94F82B
feba:					# SHA-512 start values: H5, H4, H1, H0
.quad	0x9B05688C2B3E6C1F, 0x510E527FADE682D1, 0xBB67AE8584CAA73B, 0x6A09E667F3BCC908
k512:					# SHA-512 round constants
.quad	0x428A2F98D728AE22, 0x7137449123EF65CD, 0xB5C0FBCFEC4D3B2F, 0xE9B5DBA58189DBBC
.quad	0x3956C25BF348B538, 0x59F111F1B605D019, 0x923F82A4AF194F9B, 0xAB1C5ED5DA6D8118
.quad	0xD807AA98A3030242, 0x12835B0145706FBE, 0x243185BE4EE4B28C, 0x550C7DC3D5FFB4E2
.quad	0x72BE5D74F27B896F, 0x80DEB1FE3B1696B1, 0x9BDC06A725C71235, 0xC19BF174CF692694
.quad	0xE49B69C19EF14AD2, 0xEFBE4786384F25E3, 0x0FC19DC68B8CD5B5, 0x240CA1CC77AC9C65
.quad	0x2DE92C6F592B0275, 0x4A7484AA6EA6E483, 0x5CB0A9DCBD41FBD4, 0x76F988DA831153B5
.quad	0x983E5152EE66DFAB, 0xA831C66D2DB43210, 0xB00327C898FB213F, 0xBF597FC7BEEF0EE4
.quad	0xC6E00BF33DA88FC2, 0xD5A79147930AA725, 0x06CA6351E003826F, 0x142929670A0E6E70
.quad	0x27B70A8546D22FFC, 0x2E1B21385C26C926, 0x4D2C6DFC5AC42AED, 0x53380D139D95B3DF
.quad	0x650A73548BAF63DE, 0x766A0ABB3C77B2A8, 0x81C2C92E47EDAEE6, 0x92722C851482353B
.quad	0xA2BFE8A14CF10364, 0xA81A664BBC423001, 0xC24B8B70D0F89791, 0xC76C51A30654BE30
.quad	0xD192E819D6EF5218, 0xD69906245565A910, 0xF40E35855771202A, 0x106AA07032BBD1B8
.quad	0x19A4C116B8D2D0C8, 0x1E376C085141AB53, 0x2748774CDF8EEB99, 0x34B0BCB5E19B48A8
.quad	0x391C0CB3C5C95A63, 0x4ED8AA4AE3418ACB, 0x5B9CCA4F7763E373, 0x682E6FF3D6B2B8A3
.quad	0x748F82EE5DEFB2FC, 0x78A5636F43172F60, 0x84C87814A1F0AB72, 0x8CC702081A6439EC
.quad	0x90BEFFFA23631E28, 0xA4506CEBDE82BDE9, 0xBEF9A3F7B2C67915, 0xC67178F2E372532B
.quad	0xCA273ECEEA26619C, 0xD186B8C721C0C207, 0xEADA7DD6CDE0EB1E, 0xF57D4F7FEE6ED178
.quad	0x06F067AA72176FBA, 0x0A637DC5A2C898A6, 0x113F9804BEF90DAE, 0x1B710B35131C471B
.quad	0x28DB77F523047D84, 0x32CAAB7B40C72493, 0x3C9EBE0A15C9BEBC, 0x431D67C49C100D4C
.quad	0x4CC5D4BECB3E42B6, 0x597F299CFC657E2A, 0x5FCB6FAB3AD6FAEC, 0x6C44198C4A475817
.size	k512, .-k512
.type	k512, @object

.macro	sha512	m0, m1, m2, m3		# 2*2 rounds
.if \@ < 4
	vmovdqu	block+16*\@(%rdi), \m0
	vpshufb	%ymm3, \m0, \m0		# \m0 = {ntohll(block[\@*16]), ntohll(block[\@*16+1]),
					#        ntohll(block[\@*16+2]), ntohll(block[\@*16+3])}
					#     = {w[\@*16], w[\@*16+1], w[\@*16+2], w[\@*16+3]}
.else
					# \m0 = {w[\@*16-16], w[\@*16-15], w[\@*16-14], w[\@*16-13]},
					# \m1 = {w[\@*16-12], w[\@*16-11], w[\@*16-10], w[\@*16-9]},
					# \m2 = {w[\@*16-8],  w[\@*16-7],  w[\@*16-6],  w[\@*16-5]},
					# \m3 = {w[\@*16-4],  w[\@*16-3],  w[\@*16-2],  w[\@*16-1]}
	vsha512msg1 \m1, \m0
	vpblendd $3, \m3, \m2, %ymm0	# ymm0 = {w[\@*16-4], w[\@*16-7], w[\@*16-6], w[\@*16-5]}
	vpermq	$57, %ymm0, %ymm0	# ymm0 = {w[\@*16-7], w[\@*16-6], w[\@*16-5], w[\@*16-4]}
	vpaddq	%ymm0, \m0, \m0
	vsha512msg2 \m3, \m0		# \m0 = {w[\@*16], w[\@*16+1], w[\@*16+2], w[\@*16+3]}
.endif
	vpaddq	\@%4*32(%r11), \m0, %ymm0
					# ymm0 = {(w+k)[\@*16], (w+k)[\@*16+1], (w+k)[\@*16+2], (w+k)[\@*16+3]}
	vsha512rnds2 %ymm0, %ymm1, %ymm2
					# ymm2 = {f', e', b', a'},
					# ymm1 = {h', g', d', c'}
.if 1
	vperm2i128 $129, %ymm0, %ymm0, %ymm0
					# ymm0 = {(w+k)[\@*16+2], (w+k)[\@*16+3], 0, 0}
.elseif 1
	vextracti128 $1, %ymm0, %xmm0	# ymm0 = {(w+k)[\@*16+2], (w+k)[\@*16+3], 0, 0}
.else
	vpermq	$78, %ymm0, %ymm0	# ymm0 = {(w+k)[\@*16+2], (w+k)[\@*16+3], (w+k)[\@*16+2], (w+k)[\@*16+3]}
.endif
	vsha512rnds2 %ymm0, %ymm2, %ymm1
					# ymm1 = {f", e", b", a"},
					# ymm2 = {h", g", d", c"}
.endm
.text

sha512_core:				# void SHA512_Core(SHA512_CTX *context)

	vmovdqu	state+32(%rdi), %ymm1	# ymm1 = {f, e, b, a}
	vmovdqu	state(%rdi), %ymm2	# ymm2 = {h, g, d, c}
	vmovdqa	endian, %ymm3		# ymm3 = constant for endian conversion
	leaq	k512, %r11
.rept 5					# 5*4*4 rounds
	sha512	%ymm4, %ymm5, %ymm6, %ymm7
	sha512	%ymm5, %ymm6, %ymm7, %ymm4
	sha512	%ymm6, %ymm7, %ymm4, %ymm5
	sha512	%ymm7, %ymm4, %ymm5, %ymm6
	subq	$-128, %r11
.endr
	vmovdqu	state+32(%rdi), %ymm7	# ymm7 = {f, e, b, a}
	vmovdqu	state(%rdi), %ymm6	# ymm6 = {h, g, d, c}
	vpaddq	%ymm7, %ymm1, %ymm1	# ymm1 = {f', e', b', a'} + {f, e, b, a}
					#      = {f", e", b", a"}
	vpaddq	%ymm6, %ymm2, %ymm2	# ymm2 = {h', g', d', c'} + {h, g, d, c}
					#      = {h", g", d", c"}
	vmovdqu	%ymm1, state+32(%rdi)
	vmovdqu	%ymm2, state(%rdi)	# state = {h", g", d", c", f", e", b", a"}
	retq

.global	sha512_core
.size	sha512_core, .-sha512_core
.type	sha512_core, @function

sha512_final:				# void SHA512_Final(SHA512_CTX    *context,
					#                   unsigned char digest[64])

	movq	%rdi, %rdx		# rdx = address of context structure
	movq	count(%rdi), %rcx	# rcx = low double word of count
	andl	$127, %ecx		# rcx = number of bytes in block
					#     = index of first free byte in block
	leaq	block(%rdx, %rcx), %rdi	# rdi = address of first free byte in block
	movq	%rdi, %r11
.Lpad_1:
	movb	$0b10000000, %al
	stosb				# block[index] = 0b10000000
.Lpad_0:
	xorl	%eax, %eax		# rax = 0
	xorl	$127, %ecx		# rcx = number of free bytes in block - 1
					#     = 127 - index
	rep	stosb			# block[index + 1, 127] = 0
	subq	$16, %rdi		# rdi = address of last octa word in block
	cmpq	%r11, %rdi
	ja	.Lpad_count		# space for count available in block?
					# index < 112?
	pushq	%rdi
	movq	%rdx, %rdi		# rdi = address of context structure
	callq	sha512_core
	popq	%rdi			# rdi = address of last octa word in block
.Lpad_block:
	movq	%rdi, %rcx		# rcx = address of last octa word in block
	leaq	block(%rdx), %rdi	# rdi = address of block
#	xorl	%eax, %eax		# rax = 0
	subq	%rdi, %rcx		# rcx = number of bytes before last octa word
					#     = 112
	rep	stosb			# block[0, 111] = 0,
					# rdi = address of last octa word in block
.Lpad_count:
	movq	count(%rdx), %rcx
	movq	count+8(%rdx), %rax	# rax:rcx = count
	shldq	$3, %rcx, %rax
	shlq	$3, %rcx		# rax:rcx = count * 8
					#         = number of message bits
	bswapq	%rax
	bswapq	%rcx			# rax:rcx = htonlll(number of message bits)
	stosq
	movq	%rcx, %rax
	stosq				# block[112, 127] = htonlll(number of message bits)
	pushq	%rdi
	movq	%rdx, %rdi		# rdi = address of context structure
	callq	sha512_core
	popq	%rdi
.Ldigest:
	vmovdqa	reverse, %ymm0
	vmovdqu	state(%rdx), %ymm1	# ymm1 = {h, g, d, c}
	vmovdqu	state+32(%rdx), %ymm2	# ymm2 = {f, e, b, a}
	vpshufb	%ymm0, %ymm1, %ymm1	# ymm1 = {htonll(g), htonll(h), htonll(c), htonll(d)}
	vpshufb	%ymm0, %ymm2, %ymm2	# ymm2 = {htonll(e), htonll(f), htonll(a), htonll(b)}
	vperm2i128 $19, %ymm2, %ymm1, %ymm0
					# ymm0 = {htonll(a), htonll(b), htonll(c), htonll(d)}
	vmovdqu	%ymm0, (%rsi)
	vperm2i128 $2, %ymm2, %ymm1, %ymm0
					# ymm0 = {htonll(e), htonll(f), htonll(g), htonll(h)}
	vmovdqu	%ymm0, 32(%rsi)
	retq

.global	sha512_final
.size	sha512_final, .-sha512_final
.type	sha512_final, @function

sha512_init:				# void SHA512_Init(SHA512_CTX *context)

	pxor	%xmm0, %xmm0		# xmm0 = 0
	vmovdqa	hgdc, %ymm1		# ymm1 = {H7, H6, H3, H2}
	vmovdqa	feba, %ymm2		# ymm2 = {H5, H4, H1, H0}
	movdqu	%xmm0, count(%rdi)	# count = 0
	vmovdqu	%ymm1, state(%rdi)
	vmovdqu	%ymm2, state+32(%rdi)	# state = {H7, H6, H3, H2, H5, H4, H1, H0}
					#       = {h, g, d, c, f, e, b, a}
	retq

.global	sha512_init
.size	sha512_init, .-sha512_init
.type	sha512_init, @function

sha512_update:				# void SHA512_Update(SHA512_CTX         *context,
					#                    void const         *data,
					#                    unsigned long long size)
	testq	%rdx, %rdx
	jz	.Lnone			# no data?

	movq	count(%rdi), %rcx	# rcx = low quad word of count
	andl	$127, %ecx		# rcx = number of bytes in block
					#     = index of first free byte in block
	addq	%rdx, count(%rdi)
	adcq	$0, count+8(%rdi)	# count += number of bytes in data
	movq	%rdi, %rax		# rax = address of context structure
.Ldata:
	leaq	block(%rax, %rcx), %rdi	# rdi = address of first free byte in block
	xorl	$127, %ecx
	incl	%ecx			# rcx = number of free bytes in block
	subq	%rcx, %rdx		# rdx = number of bytes in data
					#     - number of free bytes in block
	jb	.Llast			# number of bytes in data < number of free bytes in block?
.Lmore:
	rep	movsb			# rsi = address of remaining data,
					# rcx = 0 = index of first free byte in block
	pushq	%rdi
	movq	%rax, %rdi		# rdi = address of context structure
	callq	sha512_core
	popq	%rdi
	testq	%rdx, %rdx
	jnz	.Ldata			# more data?
.Lnone:
	retq
.Llast:
	addq	%rdx, %rcx		# rcx = number of bytes in data
	rep	movsb
	retq

.global	sha512_update
.size	sha512_update, .-sha512_update
.type	sha512_update, @function
.end
+EOF
else
as --64 --defsym ALIGNED=yes --no-pad-sections --noexecstack -o sha2-512.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

# Unix System V calling convention for AMD64 platform:
# - first 6 floating-point arguments (from left to right) are passed in
#   registers XMM0 to XMM5;
# - first 6 integer or pointer arguments (from left to right) are passed
#   in registers RDI/R7, RSI/R6, RDX/R2, RCX/R1, R8 and R9
#   (R10 is used as static chain pointer in case of nested functions);
# - surplus arguments are pushed on stack in reverse order (from right to
#   left), 8-byte aligned;
# - 128-bit integer arguments are passed as pair of 64-bit integer arguments,
#   low part before/below high part;
# - 128-bit integer result is returned in registers RAX/R0 (low part) and
#   RDX/R2 (high part);
# - 64-bit integer or pointer result is returned in register RAX/R0;
# - 32-bit integer result is returned in register EAX;
# - floating-point result is returned in register XMM0;
# - registers RBX/R3, RSP/R4, RBP/R5 and R12 to R15 must be preserved;
# - registers RAX/R0, RCX/R1, RDX/R2, RSI/R6, RDI/R7, R8, R9, R10 (in
#   case of normal functions), R11 and XMM0 to XMM15 are volatile and can
#   be clobbered;
# - stack is 16-byte aligned: callee must decrement RSP by 8+n*16 bytes
#   before calling other functions (CALL instruction pushes 8 bytes);
# - a "red zone" of 128 bytes below the stack pointer can be clobbered.

.ident	"Copyright (C) 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>"
.file	"sha2-512.s"
.arch	generic64
.code64
.att_syntax

.struct	0				# SHA512_CTX structure
state:					# 8 quad words
.space	8*8
count:					# 1 octa word
.space	2*8
block:					# 16 quad words
.space	16*8

.altmacro				# ??? %r%(15-(\t)&7) ???
.macro	sha512	k :req, t :req
.if \t < 16
	movq	8*\t(%rsi), %rax
	bswapq	%rax			# rax = ntohll(block[t])
.else
	movq	8*((\t-15)&15)(%rsi), %rax
	movq	8*((\t-2)&15)(%rsi), %rbx
	movq	%rax, %rcx		# rax = m = block[t-15&15]
	movq	%rbx, %rdx		# rbx = n = block[t-2&15]
	shrq	$7, %rax		# rax = m >> 7
	shrq	$6, %rbx		# rbx = n >> 6
	rorq	$1, %rcx		# rcx = ROTR(m, 1)
	rorq	$19, %rdx		# rdx = ROTR(n, 19)
	xorq	%rcx, %rax		# rax = (m >> 7) ^ ROTR(m, 1)
	xorq	%rdx, %rbx		# rbx = (n >> 6) ^ ROTR(n, 19)
	rorq	$(8-1), %rcx		# rcx = ROTR(m, 8)
	rorq	$(61-19), %rdx		# rdx = ROTR(n, 61)
	xorq	%rcx, %rax		# rax = (m >> 7) ^ ROTR(m, 1) ^ ROTR(m, 8)
					#     = SMALL_0(m)
	xorq	%rdx, %rbx		# rbx = (n >> 6) ^ ROTR(n, 19) ^ ROTR(n, 61)
					#     = SMALL_1(n)
	addq	8*((\t-16)&15)(%rsi), %rax
	addq	8*((\t-7)&15)(%rsi), %rbx
	addq	%rbx, %rax		# rax = SMALL_0(block[t-15&15]) + block[t-16&15]
					#     + SMALL_1(block[t-2&15]) + block[t-7&15]
.endif
	movq	%rax, 8*((\t)&15)(%rsi)	# block[t&15] = (t < 16)
					#             ? ntohll(block[t])
					#             : SMALL_0(block[t-15&15]) + block[t-16&15]
					#             + SMALL_1(block[t-2&15]) + block[t-7&15]
	addq	8*((87-\t)&7)(%rdi), %rax
					# rax = block[t&15] + state[71-t&7]
					#     = block[t&15] + h
	movq	8*((86-\t)&7)(%rdi), %rbx
					# rbx = g = state[70-t&7]
	movq	8*((85-\t)&7)(%rdi), %rcx
					# rcx = f = state[69-t&7]
	movq	8*((84-\t)&7)(%rdi), %rdx
					# rdx = e = state[68-t&7]
	xorq	%rbx, %rcx		# rcx = f ^ g
	andq	%rdx, %rcx		# rcx = e & (f ^ g)
	xorq	%rbx, %rcx		# rcx = e & (f ^ g) ^ g
					#     = CH(e, f, g)
	addq	%rax, %rcx		# rcx = block[t&15] + CH(e, f, g) + h
	movq	\k, %rax		# rax = k[t]
	addq	%rcx, %rax		# rax = block[t&15] + CH(e, f, g) + h + k[t]
	movq	%rdx, %rbx		# rbx = e
	rorq	$14, %rdx		# rdx = ROTR(e, 14)
	rorq	$18, %rbx		# rbx = ROTR(e, 18)
	xorq	%rbx, %rdx		# rdx = ROTR(e, 14) ^ ROTR(e, 18)
	rorq	$(41-18), %rbx		# rbx = ROTR(e, 41)
	xorq	%rdx, %rbx		# rbx = ROTR(e, 14) ^ ROTR(e, 18) ^ ROTR(e, 41)
					#     = SIGMA_1(e)
	addq	%rbx, %rax		# rax = SIGMA_1(e) + CH(e, f, g) + h + k[t] + block[t&15]
					#     = T1
	movq	8*((80-\t)&7)(%rdi), %rbx
	movq	8*((81-\t)&7)(%rdi), %rcx
					# rcx = b
	movq	%rbx, %rdx		# rdx = a
	xorq	%rcx, %rbx		# rbx = a ^ b
	andq	%rdx, %rcx		# rcx = a & b
	andq	8*((82-\t)&7)(%rdi), %rbx
					# rbx = (a ^ b) & c
	addq	%rax, 8*((83-\t)&7)(%rdi)
					# d' = d + T1
	orq	%rcx, %rbx		# rbx = (a & b) | ((a ^ b) & c)
					#     = MAJ(a, b, c)
	addq	%rbx, %rax		# rax = T1 + MAJ(a, b, c)
	movq	%rdx, %rcx		# rcx = a
	rorq	$28, %rdx		# rdx = ROTR(a, 28)
	rorq	$34, %rcx		# rcx = ROTR(a, 34)
	xorq	%rcx, %rdx		# rdx = ROTR(a, 28) ^ ROTR(a, 34)
	rorq	$(39-34), %rcx		# rcx = ROTR(a, 39)
	xorq	%rcx, %rdx		# rdx = ROTR(a, 28) ^ ROTR(a, 34) ^ ROTR(a, 39)
					#     = SIGMA_0(a)
	addq	%rdx, %rax		# rax = T1 + T2
	movq	%rax, 8*((87-\t)&7)(%rdi)
					# h' = T1 + T2
.endm
.text

sha512_core:				# void SHA512_Core(SHA512_CTX *context)

	pushq	%rbx
	pushq	%r12
	pushq	%r13
	pushq	%r14
	pushq	%r15
					# load working variables from state
	movq	state(%rdi), %r8	# r8 = a
	movq	state+8(%rdi), %r9	# r9 = b
	movq	state+16(%rdi), %r10	# r10 = c
	movq	state+24(%rdi), %r11	# r11 = d
	movq	state+32(%rdi), %r12	# r12 = e
	movq	state+40(%rdi), %r13	# r13 = f
	movq	state+48(%rdi), %r14	# r14 = g
	movq	state+56(%rdi), %r15	# r15 = h
	leaq	block(%rdi), %rsi
					# calculate 80 rounds
	sha512	0x428A2F98D728AE22, 0
	sha512	0x7137449123EF65CD, 1
	sha512	0xB5C0FBCFEC4D3B2F, 2
	sha512	0xE9B5DBA58189DBBC, 3
	sha512	0x3956C25BF348B538, 4
	sha512	0x59F111F1B605D019, 5
	sha512	0x923F82A4AF194F9B, 6
	sha512	0xAB1C5ED5DA6D8118, 7
	sha512	0xD807AA98A3030242, 8
	sha512	0x12835B0145706FBE, 9
	sha512	0x243185BE4EE4B28C, 10
	sha512	0x550C7DC3D5FFB4E2, 11
	sha512	0x72BE5D74F27B896F, 12
	sha512	0x80DEB1FE3B1696B1, 13
	sha512	0x9BDC06A725C71235, 14
	sha512	0xC19BF174CF692694, 15
	sha512	0xE49B69C19EF14AD2, 16
	sha512	0xEFBE4786384F25E3, 17
	sha512	0x0FC19DC68B8CD5B5, 18
	sha512	0x240CA1CC77AC9C65, 19
	sha512	0x2DE92C6F592B0275, 20
	sha512	0x4A7484AA6EA6E483, 21
	sha512	0x5CB0A9DCBD41FBD4, 22
	sha512	0x76F988DA831153B5, 23
	sha512	0x983E5152EE66DFAB, 24
	sha512	0xA831C66D2DB43210, 25
	sha512	0xB00327C898FB213F, 26
	sha512	0xBF597FC7BEEF0EE4, 27
	sha512	0xC6E00BF33DA88FC2, 28
	sha512	0xD5A79147930AA725, 29
	sha512	0x06CA6351E003826F, 30
	sha512	0x142929670A0E6E70, 31
	sha512	0x27B70A8546D22FFC, 32
	sha512	0x2E1B21385C26C926, 33
	sha512	0x4D2C6DFC5AC42AED, 34
	sha512	0x53380D139D95B3DF, 35
	sha512	0x650A73548BAF63DE, 36
	sha512	0x766A0ABB3C77B2A8, 37
	sha512	0x81C2C92E47EDAEE6, 38
	sha512	0x92722C851482353B, 39
	sha512	0xA2BFE8A14CF10364, 40
	sha512	0xA81A664BBC423001, 41
	sha512	0xC24B8B70D0F89791, 42
	sha512	0xC76C51A30654BE30, 43
	sha512	0xD192E819D6EF5218, 44
	sha512	0xD69906245565A910, 45
	sha512	0xF40E35855771202A, 46
	sha512	0x106AA07032BBD1B8, 47
	sha512	0x19A4C116B8D2D0C8, 48
	sha512	0x1E376C085141AB53, 49
	sha512	0x2748774CDF8EEB99, 50
	sha512	0x34B0BCB5E19B48A8, 51
	sha512	0x391C0CB3C5C95A63, 52
	sha512	0x4ED8AA4AE3418ACB, 53
	sha512	0x5B9CCA4F7763E373, 54
	sha512	0x682E6FF3D6B2B8A3, 55
	sha512	0x748F82EE5DEFB2FC, 56
	sha512	0x78A5636F43172F60, 57
	sha512	0x84C87814A1F0AB72, 58
	sha512	0x8CC702081A6439EC, 59
	sha512	0x90BEFFFA23631E28, 60
	sha512	0xA4506CEBDE82BDE9, 61
	sha512	0xBEF9A3F7B2C67915, 62
	sha512	0xC67178F2E372532B, 63
	sha512	0xCA273ECEEA26619C, 64
	sha512	0xD186B8C721C0C207, 65
	sha512	0xEADA7DD6CDE0EB1E, 66
	sha512	0xF57D4F7FEE6ED178, 67
	sha512	0x06F067AA72176FBA, 68
	sha512	0x0A637DC5A2C898A6, 69
	sha512	0x113F9804BEF90DAE, 70
	sha512	0x1B710B35131C471B, 71
	sha512	0x28DB77F523047D84, 72
	sha512	0x32CAAB7B40C72493, 73
	sha512	0x3C9EBE0A15C9BEBC, 74
	sha512	0x431D67C49C100D4C, 75
	sha512	0x4CC5D4BECB3E42B6, 76
	sha512	0x597F299CFC657E2A, 77
	sha512	0x5FCB6FAB3AD6FAEC, 78
	sha512	0x6C44198C4A475817, 79
					# add working variables to state
	addq	%r8, state(%rdi)
	addq	%r9, state+8(%rdi)
	addq	%r10, state+16(%rdi)
	addq	%r11, state+24(%rdi)
	addq	%r12, state+32(%rdi)
	addq	%r13, state+40(%rdi)
	addq	%r14, state+48(%rdi)
	addq	%r15, state+56(%rdi)

	popq	%r15
	popq	%r14
	popq	%r13
	popq	%r12
	popq	%rbx
	retq

.global	sha512_core
.size	sha512_core, .-sha512_core
.type	sha512_core, @function

sha512_final:				# void SHA512_Final(SHA512_CTX    *context,
					#                   unsigned char digest[64])

	movq	%rdi, %rdx		# rdx = address of context structure
	movq	count(%rdi), %rcx	# rcx = low quad word of count
	andl	$127, %ecx		# rcx = number of bytes in block
					#     = index of first free byte in block
	leaq	block(%rdi, %rcx), %rdi	# rdi = address of first free byte in block
	movq	%rdi, %r11
.Lpad_1:
	movb	$0b10000000, %al
	stosb				# block[index] = 0b10000000
.Lpad_0:
	xorl	%eax, %eax		# rax = 0
	xorl	$127, %ecx		# rcx = number of free bytes in block - 1
					#     = 127 - index
	rep	stosb			# block[index + 1, 127] = 0
	subq	$16, %rdi		# rdi = address of last octa word in block
	cmpq	%r11, %rdi
	ja	.Lpad_count		# space for count available in block?
					# index < 112?
	pushq	%rsi
	pushq	%rdi
	pushq	%rdx
	movq	%rdx, %rdi		# rdi = address of context structure
	callq	sha512_core
	popq	%rdx
	popq	%rdi			# rdi = address of last octa word in block
	popq	%rsi			# rsi = address of digest
.Lpad_block:
	movq	%rdi, %rcx		# rcx = address of last octa word in block
	leaq	block(%rdx), %rdi	# rdi = address of block
	xorl	%eax, %eax		# rax = 0
	subq	%rdi, %rcx		# rcx = number of bytes before last octa word
					#     = 112
	rep	stosb			# block[0, 111] = 0,
					# rdi = address of last octa word in block
.Lpad_count:
	movq	count(%rdx), %rcx
	movq	count+8(%rdx), %rax	# rax:rcx = count
	shldq	$3, %rcx, %rax
	shlq	$3, %rcx		# rax:rcx = count * 8
					#         = number of message bits
	bswapq	%rax
	bswapq	%rcx			# rax:rcx = htonlll(number of message bits)
	stosq
	movq	%rcx, %rax
	stosq				# block[112, 127] = htonlll(number of message bits)
	pushq	%rsi
	pushq	%rdi
	pushq	%rdx
	movq	%rdx, %rdi		# rdi = address of context structure
	callq	sha512_core
	popq	%rsi			# rsi = address of state
	popq	%rax
	popq	%rdi			# rdi = address of digest
	movl	$8, %ecx		# ecx = number of quad words
.Ldigest:
	lodsq
.ifdef ALIGNED
	bswapq	%rax
	stosq
.else
	rolq	$8, %rax
	stosb
	rolq	$8, %rax
	stosb
	rolq	$8, %rax
	stosb
	rolq	$8, %rax
	stosb
	rolq	$8, %rax
	stosb
	rolq	$8, %rax
	stosb
	rolq	$8, %rax
	stosb
	rolq	$8, %rax
	stosb
.endif
	decl	%ecx
	jnz	.Ldigest

	retq

.global	sha512_final
.size	sha512_final, .-sha512_final
.type	sha512_final, @function

sha512_init:				# void SHA512_Init(SHA512_CTX *context)

	movq	0x6A09E667F3BCC908, %rax
	stosq				# state[0] = H0
	movq	0xBB67AE8584CAA73B, %rax
	stosq				# state[1] = H1
	movq	0x3C6EF372FE94F82B, %rax
	stosq				# state[2] = H2
	movq	0xA54FF53A5F1D36F1, %rax
	stosq				# state[3] = H3
	movq	0x510E527FADE682D1, %rax
	stosq				# state[4] = H4
	movq	0x9B05688C2B3E6C1F, %rax
	stosq				# state[5] = H5
	movq	0x1F83D9ABFB41BD6B, %rax
	stosq				# state[6] = H6
	movq	0x5BE0CD19137E2179, %rax
	stosq				# state[7] = H7
	xorl	%eax, %eax
	stosq
	stosq				# count = 0
	retq

.global	sha512_init
.size	sha512_init, .-sha512_init
.type	sha512_init, @function

sha512_update:				# void SHA512_Update(SHA512_CTX         *context,
					#                    void const         *data,
					#                    unsigned long long size)
	testq	%rdx, %rdx
	jz	.Lnone			# no data?

	movq	count(%rdi), %rcx	# rcx = low quad word of count
	andl	$127, %ecx		# rcx = number of bytes in block
					#     = index of first free byte in block
	addq	%rdx, count(%rdi)
	adcq	$0, count+8(%rdi)	# count += number of bytes in data
	movq	%rdi, %rax		# rax = address of context structure
.Ldata:
	leaq	block(%rax, %rcx), %rdi	# rdi = address of first free byte in block
	xorl	$127, %ecx
	incl	%ecx			# rcx = number of free bytes in block
	subq	%rcx, %rdx		# rdx = number of bytes in data
					#     - number of free bytes in block
	jb	.Llast			# number of bytes in data < number of free bytes in block?
.Lmore:
	rep	movsb			# rsi = address of remaining data
	pushq	%rsi
	pushq	%rdx
	pushq	%rax
	movq	%rax, %rdi		# rdi = address of context structure
	callq	sha512_core
	popq	%rax
	popq	%rdx
	popq	%rsi
	xorl	%ecx, %ecx		# rcx = 0 = index of first free byte in block
	testq	%rdx, %rdx
	jnz	.Ldata			# more data?
.Lnone:
	retq
.Llast:
	addq	%rdx, %rcx		# rcx = number of bytes in data
	rep	movsb
	retq

.global	sha512_update
.size	sha512_update, .-sha512_update
.type	sha512_update, @function
.end
+EOF
fi

gcc -m64 -O2 -x c - -x none sha2-512.o <<+\E\O\F && ./a.out
// Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

typedef struct _sha512_ctx {
    unsigned long long state[8], count[2], block[16];
} sha512_ctx;

void sha512_core(sha512_ctx *context);
void sha512_final(sha512_ctx *context, unsigned char digest[64]);
void sha512_init(sha512_ctx *context);
void sha512_update(sha512_ctx *context, unsigned char const data[], unsigned long long size);

#include <arpa/inet.h>          // for ntohl() function
#include <stdio.h>
#include <string.h>             // for memcpy() and memset() functions
#include <time.h>

unsigned long long ntohll(unsigned long long n)
{
#if 0
    unsigned int       i = 0;
    unsigned long long h = 0;

    do {
        h <<= 8;
        h |= ((unsigned char *) &n)[i];
    } while (++i < 8);

    return h;
#elif 0
    return ((unsigned char *) &n)[0] * 0x0100000000000000u
         | ((unsigned char *) &n)[1] * 0x0001000000000000u
         | ((unsigned char *) &n)[2] * 0x0000010000000000u
         | ((unsigned char *) &n)[3] * 0x0000000100000000u
         | ((unsigned char *) &n)[4] * 0x0000000001000000u
         | ((unsigned char *) &n)[5] * 0x0000000000010000u
         | ((unsigned char *) &n)[6] * 0x0000000000000100u
         | ((unsigned char *) &n)[7];
#else
    return ntohl(1) == 1 ? n : ntohl((unsigned long) (n >> 32))
                             | ntohl((unsigned long) n) * 0x100000000u;
#endif
}

int main()
{
    unsigned char digest[64], million[1000000];
    unsigned int  n = 1024 * 1024 * 1024 / 128;

    clock_t t;

    sha512_ctx context;

    sha512_init(&context);
    sha512_final(&context, digest);

    printf("\"\"\n"
           "\tcf83e1357eefb8bd f1542850d66d8007 d620e4050b5715dc 83f4a921d36ce9ce\n"
           "\t%016llX %016llX %016llX %016llX\n\t%016llX %016llX %016llX %016llX\n"
           "\t47d0d13c5d85f2b0 ff8318d2877eec2f 63b931bd47417a81 a538327af927da3e\n",
           ntohll(context.state[0]), ntohll(context.state[1]),
           ntohll(context.state[2]), ntohll(context.state[3]),
           ntohll(context.state[4]), ntohll(context.state[5]),
           ntohll(context.state[6]), ntohll(context.state[7]));

    sha512_init(&context);
    sha512_update(&context, "abc", 3);
    sha512_final(&context, digest);

    printf("\"abc\"\n"
           "\tddaf35a193617aba cc417349ae204131 12e6fa4e89a97ea2 0a9eeee64b55d39a\n"
           "\t%016llX %016llX %016llX %016llX\n\t%016llX %016llX %016llX %016llX\n"
           "\t2192992a274fc1a8 36ba3c23a3feebbd 454d4423643ce80e 2a9ac94fa54ca49f\n",
           ntohll(context.state[0]), ntohll(context.state[1]),
           ntohll(context.state[2]), ntohll(context.state[3]),
           ntohll(context.state[4]), ntohll(context.state[5]),
           ntohll(context.state[6]), ntohll(context.state[7]));

    sha512_init(&context);
    sha512_update(&context, "abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu",
                     sizeof("abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu") - 1);
    sha512_final(&context, digest);

    printf("\"abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu\"\n"
           "\t8e959b75dae313da 8cf4f72814fc143f 8f7779c6eb9f7fa1 7299aeadb6889018\n"
           "\t%016llX %016llX %016llX %016llX\n\t%016llX %016llX %016llX %016llX\n"
           "\t501d289e4900f7e4 331b99dec4b5433a c7d329eeb6dd2654 5e96e55b874be909\n",
           ntohll(context.state[0]), ntohll(context.state[1]),
           ntohll(context.state[2]), ntohll(context.state[3]),
           ntohll(context.state[4]), ntohll(context.state[5]),
           ntohll(context.state[6]), ntohll(context.state[7]));

    memset(million, 0, 1000);

    sha512_init(&context);
    sha512_update(&context, million, 111);
    sha512_final(&context, digest);

    printf("'\\0'*111\n"
           "\t77ddd3a542e530fd 047b8977c657ba6c e72f1492e360b2b2 212cd264e75ec038\n"
           "\t%016llX %016llX %016llX %016llX\n\t%016llX %016llX %016llX %016llX\n"
           "\t82e4ff0525517ab4 207d14c70c2259ba 88d4d335ee0e7e20 543d22102ab1788c\n",
           ntohll(context.state[0]), ntohll(context.state[1]),
           ntohll(context.state[2]), ntohll(context.state[3]),
           ntohll(context.state[4]), ntohll(context.state[5]),
           ntohll(context.state[6]), ntohll(context.state[7]));

    sha512_init(&context);
    sha512_update(&context, million, 112);
    sha512_final(&context, digest);

    printf("'\\0'*112\n"
           "\t2be2e788c8a8adea a9c89a7f78904cac ea6e39297d75e057 3a73c756234534d6\n"
           "\t%016llX %016llX %016llX %016llX\n\t%016llX %016llX %016llX %016llX\n"
           "\t627ab4156b48a665 7b29ab8beb733340 40ad39ead81446bb 09c70704ec707952\n",
           ntohll(context.state[0]), ntohll(context.state[1]),
           ntohll(context.state[2]), ntohll(context.state[3]),
           ntohll(context.state[4]), ntohll(context.state[5]),
           ntohll(context.state[6]), ntohll(context.state[7]));

    sha512_init(&context);
    sha512_update(&context, million, 113);
    sha512_final(&context, digest);

    printf("'\\0'*113\n"
           "\t0e67910bcf0f9ccd e5464c63b9c850a1 2a759227d16b040d 98986d54253f9f34\n"
           "\t%016llX %016llX %016llX %016llX\n\t%016llX %016llX %016llX %016llX\n"
           "\t322318e56b8feb86 c5fb2270ed87f312 52f7f68493ee7597 43909bd75e4bb544\n",
           ntohll(context.state[0]), ntohll(context.state[1]),
           ntohll(context.state[2]), ntohll(context.state[3]),
           ntohll(context.state[4]), ntohll(context.state[5]),
           ntohll(context.state[6]), ntohll(context.state[7]));

    sha512_init(&context);
    sha512_update(&context, million, 122);
    sha512_final(&context, digest);

    printf("'\\0'*122\n"
           "\t4f3f095d015be4a7 a7cc0b8c04da4aa0 9e74351e3a97651f 744c23716ebd9b3e\n"
           "\t%016llX %016llX %016llX %016llX\n\t%016llX %016llX %016llX %016llX\n"
           "\t822e5077a01baa5c c0ed45b9249e88ab 343d4333539df21e d229da6f4a514e0f\n",
           ntohll(context.state[0]), ntohll(context.state[1]),
           ntohll(context.state[2]), ntohll(context.state[3]),
           ntohll(context.state[4]), ntohll(context.state[5]),
           ntohll(context.state[6]), ntohll(context.state[7]));

    sha512_init(&context);
    sha512_update(&context, million, 1000);
    sha512_final(&context, digest);

    printf("'\\0'*1000\n"
           "\tca3dff61bb23477a a6087b27508264a6 f9126ee3a004f53c b8db942ed345f2f2\n"
           "\t%016llX %016llX %016llX %016llX\n\t%016llX %016llX %016llX %016llX\n"
           "\td229b4b59c859220 a1cf1913f34248e3 803bab650e849a3d 9a709edc09ae4a76\n",
           ntohll(context.state[0]), ntohll(context.state[1]),
           ntohll(context.state[2]), ntohll(context.state[3]),
           ntohll(context.state[4]), ntohll(context.state[5]),
           ntohll(context.state[6]), ntohll(context.state[7]));

    memset(million, 'A', 1000);

    sha512_init(&context);
    sha512_update(&context, million, 1000);
    sha512_final(&context, digest);

    printf("'A'*1000\n"
           "\t329c52ac62d1fe73 1151f2b895a00475 445ef74f50b979c6 f7bb7cae349328c1\n"
           "\t%016llX %016llX %016llX %016llX\n\t%016llX %016llX %016llX %016llX\n"
           "\td4cb4f7261a0ab43 f936a24b000651d4 a824fcdd577f211a ef8f806b16afe8af\n",
           ntohll(context.state[0]), ntohll(context.state[1]),
           ntohll(context.state[2]), ntohll(context.state[3]),
           ntohll(context.state[4]), ntohll(context.state[5]),
           ntohll(context.state[6]), ntohll(context.state[7]));

    memset(million, 'U', 1005);

    sha512_init(&context);
    sha512_update(&context, million, 1005);
    sha512_final(&context, digest);

    printf("'U'*1005\n"
           "\t59f5e54fe299c6a8 764c6b199e44924a 37f59e2b56c3ebad 939b7289210dc8e4\n"
           "\t%016llX %016llX %016llX %016llX\n\t%016llX %016llX %016llX %016llX\n"
           "\tc21b9720165b0f4d 4374c90f1bf4fb4a 5ace17a116179801 5052893a48c3d161\n",
           ntohll(context.state[0]), ntohll(context.state[1]),
           ntohll(context.state[2]), ntohll(context.state[3]),
           ntohll(context.state[4]), ntohll(context.state[5]),
           ntohll(context.state[6]), ntohll(context.state[7]));

    memset(million, 'a', 1000000);

    sha512_init(&context);
    sha512_update(&context, million, 1000000);
    sha512_final(&context, digest);

    printf("'a'*1000000\n"
           "\tce044bc9fd43269d 5bbc946cbebc3bb7 11341115cc4abdf2 edbc3ff2c57ad4b1\n"
           "\t%016llX %016llX %016llX %016llX\n\t%016llX %016llX %016llX %016llX\n"
           "\t5deb699bda257fea 5aef9c6e55fcf4cf 9dc25a8c3ce25f2e fe90908379bff7ed\n",
           ntohll(context.state[0]), ntohll(context.state[1]),
           ntohll(context.state[2]), ntohll(context.state[3]),
           ntohll(context.state[4]), ntohll(context.state[5]),
           ntohll(context.state[6]), ntohll(context.state[7]));

    memset(million, 0, 1000000);

    sha512_init(&context);
    sha512_update(&context, million, 1000000);
    sha512_final(&context, digest);

    printf("'\\0'*1000000\n"
           "\tce044bc9fd43269d 5bbc946cbebc3bb7 11341115cc4abdf2 edbc3ff2c57ad4b1\n"
           "\t%016llX %016llX %016llX %016llX\n\t%016llX %016llX %016llX %016llX\n"
           "\t5deb699bda257fea 5aef9c6e55fcf4cf 9dc25a8c3ce25f2e fe90908379bff7ed\n",
           ntohll(context.state[0]), ntohll(context.state[1]),
           ntohll(context.state[2]), ntohll(context.state[3]),
           ntohll(context.state[4]), ntohll(context.state[5]),
           ntohll(context.state[6]), ntohll(context.state[7]));

    t = clock();

    do sha512_core(&context); while (--n);

    t = clock() - t;

    printf("%lu.%06lu seconds per GiB\n",
           t / CLOCKS_PER_SEC, (t % CLOCKS_PER_SEC) * 1000000u / CLOCKS_PER_SEC);
}
+EOF

rm a.out
